from typing import Optional, Type

from django.core.exceptions import ImproperlyConfigured
from rest_framework import generics
from rest_framework.serializers import BaseSerializer

from .mixins import (
    APIErrorsMixin,
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)


class GenericAPIView(APIErrorsMixin, generics.GenericAPIView):
    request_serializer_class: Optional[Type[BaseSerializer]] = None
    destroy_request_serializer_class: Optional[Type[BaseSerializer]] = None
    response_serializer_class: Optional[Type[BaseSerializer]] = None

    def get_request_serializer(self, *args, **kwargs) -> BaseSerializer:
        serializer = self.get_request_serializer_or_none(*args, **kwargs)
        if serializer is None:
            self.raise_request_serializer_error()
        return serializer

    def get_request_serializer_or_none(
        self, *args, **kwargs
    ) -> Optional[BaseSerializer]:
        serializer_class = self.get_request_serializer_class_or_none()
        if serializer_class is not None:
            kwargs.setdefault('context', self.get_request_serializer_context())
            return serializer_class(*args, **kwargs)

    def get_request_serializer_class_or_none(self) -> Optional[Type[BaseSerializer]]:
        if self.request.method == 'DELETE':
            return self.destroy_request_serializer_class
        return self.request_serializer_class

    def get_request_serializer_context(self):
        return self.get_serializer_context()

    def raise_request_serializer_error(self):
        raise ImproperlyConfigured(
            f'{self.__class__.__name__} should properly configure `request_serializer_class` attribute'
        )

    def get_response_serializer(self, *args, **kwargs) -> BaseSerializer:
        serializer = self.get_response_serializer_or_none(*args, **kwargs)
        if serializer is None:
            self.raise_response_serializer_error()
        return serializer

    def get_response_serializer_or_none(
        self, *args, **kwargs
    ) -> Optional[BaseSerializer]:
        serializer_class = self.get_response_serializer_class_or_none()
        if serializer_class is not None:
            kwargs.setdefault('context', self.get_response_serializer_context())
            return serializer_class(*args, **kwargs)

    def get_response_serializer_class_or_none(self) -> Optional[Type[BaseSerializer]]:
        return self.response_serializer_class

    def get_response_serializer_context(self):
        return self.get_serializer_context()

    def raise_response_serializer_error(self):
        raise ImproperlyConfigured(
            f'{self.__class__.__name__} should properly configure `response_serializer_class` attribute'
        )

    def get_serializer_class(self) -> Type[BaseSerializer]:
        response_serializer_class = self.get_response_serializer_class_or_none()
        if response_serializer_class is not None:
            return response_serializer_class

        if self.serializer_class is not None:
            return self.serializer_class

        self.raise_serializer_error()

    def raise_serializer_error(self):
        raise ImproperlyConfigured(
            f'{self.__class__.__name__} should properly configure one of these attributes: '
            f'`response_serializer_class`, `serializer_class`'
        )


# Concrete view classes that provide method handlers
# by composing the mixin classes with the base view.


class CreateAPIView(CreateModelMixin, GenericAPIView):
    """
    Concrete view for creating a model instance.
    """

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class ListAPIView(ListModelMixin, GenericAPIView):
    """
    Concrete view for listing a queryset.
    """

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)


class RetrieveAPIView(RetrieveModelMixin, GenericAPIView):
    """
    Concrete view for retrieving a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)


class DestroyAPIView(DestroyModelMixin, GenericAPIView):
    """
    Concrete view for deleting a model instance.
    """

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)


class UpdateAPIView(UpdateModelMixin, GenericAPIView):
    """
    Concrete view for updating a model instance.
    """

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class ListCreateAPIView(ListModelMixin, CreateModelMixin, GenericAPIView):
    """
    Concrete view for listing a queryset or creating a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class RetrieveUpdateAPIView(RetrieveModelMixin, UpdateModelMixin, GenericAPIView):
    """
    Concrete view for retrieving, updating a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class RetrieveDestroyAPIView(RetrieveModelMixin, DestroyModelMixin, GenericAPIView):
    """
    Concrete view for retrieving or deleting a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)


class RetrieveUpdateDestroyAPIView(
    RetrieveModelMixin, UpdateModelMixin, DestroyModelMixin, GenericAPIView
):
    """
    Concrete view for retrieving, updating or deleting a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)



from rest_framework.views import exception_handler

from .errors_formatter import ErrorsFormatter


def errors_formatter_exception_handler(exc, context):
    response = exception_handler(exc, context)

    # If unexpected error occurs (server error, etc.)
    if response is None:
        return response

    formatter = ErrorsFormatter(exc)

    response.data = formatter()

    return response



from rest_framework import exceptions
from rest_framework.settings import api_settings


class ErrorsFormatter:
    """
    Mostly copied from https://github.com/HackSoftware/Django-Styleguide
    TODO: Refactor later

    The current formatter gets invalid serializer errors,
    uses DRF standard for code and messaging
    and then parses it to the following format:
    {
        "errors": [
            {
                "message": "Error message",
                "code": "Some code",
                "field": "field_name"
            },
            {
                "message": "Error message",
                "code": "Some code",
                "field": "nested.field_name"
            },
            ...
        ]
    }
    """

    FIELD = 'field'
    MESSAGE = 'message'
    CODE = 'code'
    ERRORS = 'errors'

    def __init__(self, exception):
        self.exception = exception

    def __call__(self):
        if hasattr(self.exception, 'get_full_details'):
            formatted_errors = self._get_response_json_from_drf_errors(
                serializer_errors=self.exception.get_full_details()
            )
        else:
            formatted_errors = self._get_response_json_from_error_message(
                message=str(self.exception)
            )

        return formatted_errors

    def get_field_name(self, field_name):
        """
        Override this method if you want to change a field name returned in the response.
        For example, convert snake_case field name to camelCase.
        """
        return field_name

    def _get_response_json_from_drf_errors(self, serializer_errors=None):
        if serializer_errors is None:
            serializer_errors = {}

        if type(serializer_errors) is list:
            serializer_errors = {api_settings.NON_FIELD_ERRORS_KEY: serializer_errors}

        list_of_errors = self._get_list_of_errors(errors_dict=serializer_errors)

        response_data = {self.ERRORS: list_of_errors}

        return response_data

    def _get_response_json_from_error_message(self, *, message='', code='error'):
        response_data = {self.ERRORS: [{self.MESSAGE: message, self.CODE: code}]}

        return response_data

    def _unpack(self, obj):
        if type(obj) is list and len(obj) == 1:
            return obj[0]

        return obj

    def _get_list_of_errors(self, field_path='', errors_dict=None):
        """
        Error_dict is in the following format:
        {
            'field1': {
                'message': 'some message..'
                'code' 'some code...'
            },
            'field2: ...'
        }
        """
        if errors_dict is None:
            return []

        message_value = errors_dict.get(self.MESSAGE, None)

        # Note: If 'message' is name of a field we don't want to stop the recursion here!
        if message_value is not None and (
            type(message_value) in {str, exceptions.ErrorDetail}
        ):
            if field_path:
                errors_dict[self.FIELD] = field_path
            return [errors_dict]

        errors_list = []
        for key, value in errors_dict.items():
            new_field_path = (
                '{0}.{1}'.format(field_path, self.get_field_name(key))
                if field_path
                else self.get_field_name(key)
            )
            key_is_non_field_errors = key == api_settings.NON_FIELD_ERRORS_KEY

            if type(value) is list:
                current_level_error_list = []
                new_value = value

                for index, error in enumerate(new_value):
                    # if the type of field_error is list we need to unpack it
                    field_error = self._unpack(error)

                    if self.MESSAGE in field_error:
                        if not key_is_non_field_errors:
                            field_error[self.FIELD] = new_field_path
                        current_level_error_list.append(field_error)
                    else:
                        path = '{0}[{1}]'.format(new_field_path, index)
                        current_level_error_list.extend(
                            self._get_list_of_errors(
                                field_path=path, errors_dict=field_error
                            )
                        )
            else:
                path = field_path if key_is_non_field_errors else new_field_path

                current_level_error_list = self._get_list_of_errors(
                    field_path=path, errors_dict=value
                )

            errors_list += current_level_error_list

        return errors_list



from typing import Dict, Iterable, Optional, Type, Union

from django.core.exceptions import ImproperlyConfigured
from rest_framework import viewsets
from rest_framework.permissions import BasePermission
from rest_framework.serializers import BaseSerializer

from .generics import GenericAPIView
from .mixins import (
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)


class GenericViewSet(viewsets.ViewSetMixin, GenericAPIView):
    action_permission_classes: Optional[
        Dict[str, Union[Type[BasePermission], Iterable[Type[BasePermission]]]]
    ] = None
    request_action_serializer_classes: Optional[Dict[str, Type[BaseSerializer]]] = None
    response_action_serializer_classes: Optional[Dict[str, Type[BaseSerializer]]] = None

    def get_permission_classes_or_none(self):
        if self.action_permission_classes:
            permission_classes = self.action_permission_classes.get(self.action)
            if permission_classes is None and self.action == 'partial_update':
                permission_classes = self.action_permission_classes.get('update')
            return permission_classes

    def get_permissions(self):
        permissions = super().get_permissions()

        permission_classes = self.get_permission_classes_or_none()
        if permission_classes is not None:
            if isinstance(permission_classes, Iterable):
                for permission_class in permission_classes:
                    permissions.append(permission_class())
            else:
                permissions.append(permission_classes())

        return permissions

    def get_request_serializer_class_or_none(self) -> Optional[Type[BaseSerializer]]:
        serializer_class = None

        if self.request_action_serializer_classes:
            serializer_class = self.request_action_serializer_classes.get(self.action)
            if serializer_class is None and self.action == 'partial_update':
                serializer_class = self.request_action_serializer_classes.get('update')

        if serializer_class is None:
            return super().get_request_serializer_class_or_none()

        return serializer_class

    def raise_request_serializer_error(self):
        raise ImproperlyConfigured(
            f'{self.__class__.__name__} should properly configure `request_action_serializer_classes` attribute'
        )

    def get_response_serializer_class_or_none(self) -> Optional[Type[BaseSerializer]]:
        serializer_class = None

        if self.response_action_serializer_classes:
            serializer_class = self.response_action_serializer_classes.get(self.action)
            if serializer_class is None and self.action == 'partial_update':
                serializer_class = self.response_action_serializer_classes.get('update')

        if serializer_class is None:
            return super().get_response_serializer_class_or_none()

        return serializer_class

    def raise_response_serializer_error(self):
        raise ImproperlyConfigured(
            f'{self.__class__.__name__} should properly configure `response_action_serializer_classes` attribute'
        )

    def raise_serializer_error(self):
        raise ImproperlyConfigured(
            f'{self.__class__.__name__} should properly configure one of these attributes: '
            f'`response_action_serializer_classes`, `serializer_class`'
        )


class ReadOnlyModelViewSet(RetrieveModelMixin, ListModelMixin, GenericViewSet):
    """
    A viewset that provides default `list()` and `retrieve()` actions.
    """

    pass


class ModelViewSet(
    CreateModelMixin,
    RetrieveModelMixin,
    ListModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
    GenericViewSet,
):
    """
    A viewset that provides default `create()`, `retrieve()`, `update()`,
    `partial_update()`, `destroy()` and `list()` actions.
    """

    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rest_batteries']

package_data = \
{'': ['*']}

install_requires = \
['django>=3.0.8,<4.0.0', 'djangorestframework>=3.11.0,<4.0.0']

setup_kwargs = {
    'name': 'django-rest-batteries',
    'version': '1.2.2',
    'description': 'Build clean APIs with DRF faster',
    'long_description': '[![Build Status](https://travis-ci.org/defineimpossible/django-rest-batteries.svg?branch=master)](https://travis-ci.org/github/defineimpossible/django-rest-batteries)\n[![Coverage](https://codecov.io/gh/defineimpossible/django-rest-batteries/branch/master/graph/badge.svg)](https://codecov.io/gh/defineimpossible/django-rest-batteries)\n\n# Django REST Framework Batteries\n\nBuild clean APIs with DRF faster.\n\n# Overview\n\nHere\'s a quick overview of what the library has at the moment:\n\n- Action-based serializers for ViewSets\n- Two serializers per request/response cycle for ViewSets and GenericAPIViews\n- Action-based permissions for ViewSets\n- Single format for all errors\n\n# Requirements\n\n- Python ≥ 3.6\n- Django (2.2, 3.0)\n- Django REST Framework (3.9, 3.10, 3.11)\n\n# Installation\n\n```bash\n$ pip install django-rest-batteries\n```\n\n# Usage\n\n## Action-based serializers for ViewSets\n\nEach action can have a separate serializer:\n\n```python\nfrom rest_batteries.mixins import RetrieveModelMixin, ListModelMixin\nfrom rest_batteries.viewsets import GenericViewSet\n...\n\nclass OrderViewSet(RetrieveModelMixin,\n                   ListModelMixin,\n                   GenericViewSet):\n    response_action_serializer_classes = {\n        \'retrieve\': OrderSerializer,\n        \'list\': OrderListSerializer,\n    }\n```\n\n## Two serializers per request/response cycle\n\nWe found that more often than not we need a separate serializer for handling request payload and a separate serializer for generating response data.\n\nHow to achieve it in ViewSet:\n\n```python\nfrom rest_batteries.mixins import CreateModelMixin, ListModelMixin\nfrom rest_batteries.viewsets import GenericViewSet\n...\n\nclass OrderViewSet(CreateModelMixin,\n                   ListModelMixin,\n                   GenericViewSet):\n    request_action_serializer_classes = {\n        \'create\': OrderCreateSerializer,\n    }\n    response_action_serializer_classes = {\n        \'create\': OrderResponseSerializer,\n        \'list\': OrderResponseSerializer,\n        \'cancel\': OrderResponseSerializer,\n    }\n```\n\nHow to achieve it in GenericAPIView:\n\n```python\nfrom rest_batteries.generics import CreateAPIView\n...\n\n\nclass OrderCreateView(CreateAPIView):\n    request_serializer_class = OrderCreateSerializer\n    response_serializer_class = OrderResponseSerializer\n```\n\n## Action-based permissions for ViewSets\n\nEach action can have a separate set of permissions:\n\n```python\nfrom rest_batteries.mixins import CreateModelMixin, UpdateModelMixin, ListModelMixin\nfrom rest_batteries.viewsets import GenericViewSet\nfrom rest_framework.permissions import AllowAny, IsAuthenticated\n...\n\nclass OrderViewSet(CreateModelMixin,\n                   UpdateModelMixin,\n                   ListModelMixin,\n                   GenericViewSet):\n    action_permission_classes = {\n        \'create\': IsAuthenticated,\n        \'update\': [IsAuthenticated, IsOrderOwner],\n        \'list\': AllowAny,\n    }\n```\n\n## Single format for all errors\n\nWe believe that having a single format for all errors is good practice. This will make the process of displaying and handling errors much simpler for clients that use your APIs.\n\nAny error always will be a JSON object with a message, code (identifier of the error), and field if the error is specific to a particular field. How your response could look like:\n\n```python\n{\n    "errors": [\n        {\n            "message": "Delete or cancel all reservations first.",\n            "code": "invalid"\n        },\n        {\n            "message": "Ensure this field has no more than 21 characters.",\n            "code": "max_length",\n            "field": "address.work_phone"\n        },\n        {\n            "message": "This email already exists",\n            "code": "unique",\n            "field": "login_email"\n        }\n    ]\n}\n```\n\nYou will not have a single format out-of-the-box after installation. You need to add an exception handler to your DRF settings:\n\n```python\nREST_FRAMEWORK = {\n    ...\n    \'EXCEPTION_HANDLER\': \'rest_batteries.exception_handlers.errors_formatter_exception_handler\',\n}\n```\n\n# Credits\n\n- [Django-Styleguide by HackSoftware](https://github.com/HackSoftware/Django-Styleguide) - inspiration\n',
    'author': 'Define Impossible',
    'author_email': 'hi@defineimpossible.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/defineimpossible/django-rest-batteries',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from django.core import exceptions as django_exceptions
from rest_framework import exceptions as rest_exceptions
from rest_framework import status
from rest_framework.response import Response


class APIErrorsMixin:
    """
    Mixin that transforms Django and Python exceptions into rest_framework ones.
    Without this mixin, they return 500 status code which is not desired.
    """

    expected_exceptions = {
        ValueError: rest_exceptions.ValidationError,
        django_exceptions.ValidationError: rest_exceptions.ValidationError,
    }

    def handle_exception(self, exc):
        if isinstance(exc, tuple(self.expected_exceptions.keys())):
            drf_exception_class = self.expected_exceptions[exc.__class__]
            drf_exception = drf_exception_class(self._get_error_message(exc))

            return super().handle_exception(drf_exception)

        return super().handle_exception(exc)

    def _get_error_message(self, exc):
        if hasattr(exc, 'message_dict'):
            return exc.message_dict

        error_msg = self._get_first_matching_attr(exc, 'message', 'messages')

        if isinstance(error_msg, list):
            error_msg = ', '.join(error_msg)

        if error_msg is None:
            error_msg = str(exc)

        return error_msg

    def _get_first_matching_attr(self, obj, *attrs, default=None):
        for attr in attrs:
            if hasattr(obj, attr):
                return getattr(obj, attr)

        return default


class CreateModelMixin:
    """
    Create a model instance.
    """

    def create(self, request, *_args, **_kwargs):
        request_serializer = self.get_request_serializer(data=request.data)
        request_serializer.is_valid(raise_exception=True)

        instance = self.perform_create(request_serializer)

        response_serializer = self.get_response_serializer(instance)
        return Response(response_serializer.data, status=status.HTTP_201_CREATED)

    def perform_create(self, serializer):
        return serializer.save()


class RetrieveModelMixin:
    """
    Retrieve a model instance.
    """

    def retrieve(self, _request, *_args, **_kwargs):
        instance = self.get_object()
        serializer = self.get_response_serializer(instance)
        return Response(serializer.data)


class ListModelMixin:
    """
    List a queryset.
    """

    def list(self, _request, *_args, **_kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_response_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_response_serializer(queryset, many=True)
        return Response(serializer.data)


class _UpdateMixin:
    def _update(self, request, *_args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        request_serializer = self.get_request_serializer(
            instance, data=request.data, partial=partial
        )
        request_serializer.is_valid(raise_exception=True)

        if partial:
            instance = self.perform_partial_update(instance, request_serializer)
        else:
            instance = self.perform_update(instance, request_serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        response_serializer = self.get_response_serializer(instance)
        return Response(response_serializer.data)

    def perform_update(self, instance, serializer):
        return serializer.save()

    def perform_partial_update(self, instance, serializer):
        return self.perform_update(instance, serializer)


class UpdateModelMixin(_UpdateMixin):
    """
    Update a model instance.
    """

    def update(self, *args, **kwargs):
        return self._update(*args, **kwargs)

    def partial_update(self, *args, **kwargs):
        kwargs['partial'] = True
        return self.update(*args, **kwargs)


class FullUpdateModelMixin(_UpdateMixin):
    """
    Fully update a model instance.
    """

    def update(self, *args, **kwargs):
        return self._update(*args, **kwargs)


class PartialUpdateModelMixin(_UpdateMixin):
    """
    Partially update a model instance.
    """

    def partial_update(self, *args, **kwargs):
        kwargs['partial'] = True
        return self._update(*args, **kwargs)


class DestroyModelMixin:
    """
    Destroy a model instance.
    """

    def destroy(self, request, *_args, **_kwargs):
        instance = self.get_object()
        serializer = self.get_request_serializer_or_none(instance, data=request.data)
        if serializer is not None:
            serializer.is_valid(raise_exception=True)
            self.perform_destroy(instance, serializer)
        else:
            self.perform_destroy(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, instance, serializer=None):
        instance.delete()
